import {produce,consume,Tag} from "../src"
import {parse} from "babylon"
import generate from "babel-generator"
import * as Kit from "../src/kit"
import * as R from "ramda"
import eagerGenerators from "../src/samples/eagerGenerators"
import joinMemExprs from "../src/samples/joinMemExprs"
import looseForOf from "../src/samples/looseForOf"
import instrumentation from "../src/samples/instrumentation"
import * as Trace from "../src/trace"

const gen = ast => generate(ast,{retainLines:false,concise:true},"").code
const pretty = R.pipe(R.invoker(0,"toString"),parse,gen)

describe("join member expression", function() {
  const run = R.pipe(
    R.invoker(0,"toString"),
    parse,
    produce,
    joinMemExprs,
    consume,
    R.prop("top"),
    gen)
  it("sample1", function() {
    expect(run(`function a() {
      const a = create(), b = create(), /*@PACK*/d = create(), e = cr()
      d.a = 10
      console.log(a.a,a.c,b.a,e.c,d.c)
      if (check(b)) {
        const b = create()
        let d = create(e)
        console.log(a.a,a.c,b.a,d)
      }
    }`)).to.equal(pretty(`function a() {
      const a = create(), b = create(), /*@PACK*/ d = create(), e = cr();
      let a$a = a && a.a;
      let a$c = a && a.c;
      let d$a = d && d.a;
      let d$c = d && d.c;
      d$a = 10;
      console.log(a$a, a$c, b.a, e.c, d$c);
      if (check(b)) {
        const b = create();
        let b$a = b && b.a;
        let d = create(e);
        console.log(a$a, a$c, b$a, d);
      }
    }`))
  })
})

describe("instrumentation", function() {
  const run = R.pipe(
    R.invoker(0,"toString"),
    parse,
    produce,
    instrumentation,
    consume,
    R.prop("top"),
    gen)
  it("sample1", function() {
    expect(run(`function a() {
      function* b() {
        console.log("1")
      }
      class C {
        vo(a,b) {
         console.log("Z")
        }
        //@NOPROF
        no(a,b) {
          console.log("Z")
        }
      }
      c(v => v + 1, 
        v => {return v + 1},
        function(v) { return v + 1 })
    }`)).to.equal(pretty(`function a() {
      return e$y$prof("a@?", "1[0]",
                      function () {
                        function b() {
                          return e$y$prof$g("a.b@?", "2[6]",
                                            function* () {
                                              console.log("1");
                                            });
                        }
                        class C {
                          vo(a,b) { return e$y$prof("a.C.vo@?", "6[8]",
                                                 function () {
                                                   console.log("Z");
                                                 });
                               }
                          //@NOPROF
                          no(a,b) { console.log("Z"); }
                        }
                        c(
                          v => {
                            return e$y$prof("a.F0@?", "14[8]",
                                            function () { return v + 1; });
                          },
                          v => { return e$y$prof("a.F1@?", "15[8]",
                                                 function () { return v + 1; }); },
                          function (v) {
                            return e$y$prof("a.F2@?", "16[8]",
                                            function () { return v + 1; }); }
                        );
                      });
    }`))
  })
  context('with `this` or `arguments`', function() {
    it("sample2", function() {
      expect(run(`function a() {
      function* b() {
        console.log("1",this,arguments)
      }
      class C {
        vo() {
         console.log("Z",this)
        }
      }
      c(v => v + this.id + arguments[0], 
        v => {return v + this.id + arguments[0]},
        function x(v) {return v + this.id + arguments[0]})
    }`)).to.equal(pretty(
      `function a() {
        return e$y$prof("a@?", "1[0]",
                        function () {
                          function b() {
                            const e$y$this = this;
                            const e$y$arguments = arguments;
                            return e$y$prof$g("a.b@?", "2[6]",
                                              function* () {
                                                console.log("1", e$y$this, e$y$arguments); });
                          }
                          class C {
                            vo() {
                              const e$y$this = this;
                              return e$y$prof("a.C.vo@?", "6[8]",
                                              function () {
                                                console.log("Z", e$y$this);
                                              });
                            }
                          }
                          c(
                            v => {
                              const e$y$this = this;
                              const e$y$arguments = arguments;
                              return e$y$prof("a.F0@?", "10[8]",
                                              function () {
                                                return v + e$y$this.id + e$y$arguments[0];
                                              });
                            },
                            v => {
                              const e$y$this = this;
                              const e$y$arguments = arguments;
                              return e$y$prof("a.F1@?", "11[8]",
                                              function () {
                                                return v + e$y$this.id + e$y$arguments[0];
                                              });
                            },
                            function x(v) {
                              const e$y$this = this;
                              const e$y$arguments = arguments;
                              return e$y$prof("a.x@?", "12[8]",
                                              function () {
                                                return v + e$y$this.id + e$y$arguments[0];
                                              });
                            });
                        });
      }`))
    })
  })
})

describe("eager generators", function() {
  const run = R.pipe(
    R.invoker(0,"toString"),
    parse,
    produce,
    eagerGenerators,
    consume,
    R.prop("top"),
    gen)
  it("sample1", function() {
    expect(run(`
      function* a() {
        function* b() {
          yield* arguments
          yield 10
        }
        yield* b()
      }
      //@LAZY
      function* a() {
        function* b() {
          yield* arguments
          yield 10
        }
        yield* b()
      }
    `)).to.equal(pretty(
      `function a() {
        return e$y$make(e$y$buf => {
          function b() {
            var e$y$arguments = arguments;
            return e$y$make(e$y$buf => {
              e$y$star(e$y$buf, e$y$arguments);
              e$y(e$y$buf, 10);
            });
          }
          e$y$star(e$y$buf, b()); });
      }
      //@LAZY
      function* a() {
        function* b() {
          yield* arguments;
          yield 10;
        }
        yield* b();
      }`))
  })
})
  
describe("extra loose for-ofs", function() {
  const run = R.pipe(
    R.invoker(0,"toString"),
    parse,
    produce,
    looseForOf,
    consume,
    R.prop("top"),
    gen)
  it("sample1", function() {
    expect(run(`function a() {
      for(const i of a) {
        zzz
      }
    }`)).to.equal(pretty(`function a() {
      {
        const _e = a;
        const _arr = e$y$arr(_e);
        if (_arr != null) {
          const _len = _arr.length;
          for (let _i = 0; _i < _len; ++_i) {
            const i = _arr[_i];
            zzz;
          }
        } else {
          const _iter = _e[Symbol.iterator]();
          for (let _i = _iter.next(); !_i.done; _i = _iter.next()) {
            const i = _i.value;
            zzz;
          }
        }
      }
    }`))
  })
  it("sample2", function() {
    expect(run(`function a() {
      for(const i of a)
        zzz
    }`)).to.equal(pretty(`function a() {
      {
        const _e = a;
        const _arr = e$y$arr(_e);
        if (_arr != null) {
          const _len = _arr.length;
          for (let _i = 0; _i < _len; ++_i) {
            const i = _arr[_i];
            zzz;
          }
        } else {
          const _iter = _e[Symbol.iterator]();
          for (let _i = _iter.next(); !_i.done; _i = _iter.next()) {
            const i = _i.value;
            zzz;
          }
        }
      }
    }`))
  })
  it("sample3", function() {
    expect(run(`function a() {
      var i
      for(i of a)
        zzz
    }`)).to.equal(pretty(`function a() {
      var i
      {
        const _e = a;
        const _arr = e$y$arr(_e);
        if (_arr != null) {
          const _len = _arr.length;
          for (let _i = 0; _i < _len; ++_i) {
            i = _arr[_i];
            zzz;
          }
        } else {
          const _iter = _e[Symbol.iterator]();
          for (let _i = _iter.next(); !_i.done; _i = _iter.next()) {
            i = _i.value;
            zzz;
          }
        }
      }
    }`))
  })
  it("sample4", function() {
    expect(run(`function a() {
      var i
      for(i of a)
        for(const j of b) {
          zzz
        }
    }`)).to.equal(pretty(`function a() {
      var i;
      {
        const _e = a;
        const _arr = e$y$arr(_e);
        if (_arr != null) {
          const _len = _arr.length;
          for (let _i = 0; _i < _len; ++_i) {
            i = _arr[_i];
            {
              const _e = b;
              const _arr = e$y$arr(_e);
              if (_arr != null) {
                const _len = _arr.length;
                for (let _i = 0; _i < _len; ++_i) {
                  const j = _arr[_i];
                  zzz;
                }
              } else {
                const _iter = _e[Symbol.iterator]();
                for (let _i = _iter.next(); !_i.done; _i = _iter.next()) {
                  const j = _i.value;
                  zzz;
                }
              }
            }
          }
        } else {
          const _iter = _e[Symbol.iterator]();
          for (let _i = _iter.next(); !_i.done; _i = _iter.next()) {
            i = _i.value;
            {
              const _e = b;
              const _arr = e$y$arr(_e);
              if (_arr != null) {
                const _len = _arr.length;
                for (let _i = 0; _i < _len; ++_i) {
                  const j = _arr[_i];
                  zzz;
                }
              } else {
                const _iter = _e[Symbol.iterator]();
                for (let _i = _iter.next(); !_i.done; _i = _iter.next()) {
                  const j = _i.value;
                  zzz;
                }
              }
            }
          }
        }
      }
    }`))
  })
  it("sample5", function() {
    expect(run(`function a() {
      var i
      loo: for(i of a)
        zzz
    }`)).to.equal(pretty(`function a() {
      var i
      {
        const _e = a;
        const _arr = e$y$arr(_e);
        if (_arr != null) {
          const _len = _arr.length;
          loo: for (let _i = 0; _i < _len; ++_i) {
            i = _arr[_i];
            zzz;
          }
        } else {
          const _iter = _e[Symbol.iterator]();
          loo: for (let _i = _iter.next(); !_i.done; _i = _iter.next()) {
            i = _i.value;
            zzz;
          }
        }
      }
    }`))
  })
  
})
